@@ -150,11 +150,14 @@ class PluginGlobals(object):
             return PluginGlobals.env_registry[arg]
 
     @staticmethod
-    def push_env(arg):
+    def push_env(arg, validate=False):
         """Push the given environment on the stack."""
         if isinstance(arg,basestring):
             if not arg in PluginGlobals.env_registry:
-                raise PluginError, "Unknown environment %r" % arg
+                if validate:
+                    raise PluginError, "Unknown environment %r" % arg
+                else:
+                    env = PluginEnvironment(arg)
             env = PluginGlobals.env_registry[arg]
         else:
             env = arg
@@ -353,12 +356,17 @@ class ExtensionPoint(object):
         #
         # Construct the interface, passing in this extension
         #
-        if len(args) == 0:
+        nargs=len(args)
+        if nargs == 0:
             raise PluginError, "Must specify interface class used in the ExtensionPoint"
         self.interface = args[0]
         self.env = [ PluginGlobals.env(self.interface.__interface_namespace__) ]
-        if len(args) > 1:
-            self.env += args[1:]
+        if nargs > 1:
+            for arg in args[1:]:
+                if isinstance(arg,basestring):
+                    self.env.append( PluginGlobals.env(arg) )
+                else:
+                    self.env.append(arg)
         self.__doc__ = 'List of services that implement `%s`' % self.interface.__name__
 
     def __iter__(self):
@@ -373,15 +381,6 @@ class ExtensionPoint(object):
         Return a set of services that match the interface of this
         extension point.
         """
-        tmp = self.extensions(all=all)
-        #
-        # Return all extensions
-        #
-        if key is None:
-            return tmp
-        #
-        # Return the key'th extension
-        #
         if type(key) in (int, long):
             #
             # Q: should this be a warning?  A user _might_ be trying
@@ -389,20 +388,7 @@ class ExtensionPoint(object):
             # likely.
             #
             raise PluginError, "Access of the n-th extension point is disallowed.  This is not well-defined, since ExtensionPoints are stored as unordered sets."
-        #
-        # Return a set of extensions that match this key.  The
-        # service_name method is used, if it has been defined.  Otherwise,
-        # the interface name is used.
-        #
-        ans = set()
-        for item in tmp:
-            try:
-                if item.service_name() == key:
-                    ans.add(item)
-            except AttributeError:    
-                if str(item.name) == str(key):
-                    ans.add(item)
-        return ans
+        return self.extensions(all=all, key=key)
 
     def service(self, key=None, all=False):
         """
@@ -411,14 +397,15 @@ class ExtensionPoint(object):
         specified key, or if multiple services match.
         """
         ans = ExtensionPoint.__call__(self, key=key, all=all)
-        if len(ans) > 1:
-            raise PluginError, "The ExtensionPoint does not have a unique service!  %d services are defined for interface %s.  (key=%s)" % (len(ans), self.interface. __name__, str(key))
+        if len(ans)== 1:
+            #
+            # There is a single service, so return it.
+            #
+            return ans.pop()
         elif len(ans) == 0:
             return None
-        #
-        # There is a single service, so return it.
-        #
-        return ans.pop()
+        else:
+            raise PluginError, "The ExtensionPoint does not have a unique service!  %d services are defined for interface %s.  (key=%s)" % (len(ans), self.interface. __name__, str(key))
 
     def __len__(self):
         """
@@ -427,14 +414,14 @@ class ExtensionPoint(object):
         """
         return len(self.extensions())
 
-    def extensions(self, all=False):
+    def extensions(self, all=False, key=None):
         """
         Return a set of services that match the interface of this
         extension point.  This tacitly filters out disabled extension points.
         """
         ans = set()
         for env in self.env:
-            ans.update(env.active_services(self.interface, all=all))
+            ans.update(env.active_services(self.interface, all=all, key=key))
         return ans
 
     def __repr__(self):
@@ -477,6 +464,7 @@ class PluginEnvironment(object):
         self.log = logger_factory(self.name)
         self.log.debug("Creating PluginEnvironment %r" % self.name)
         self.level = []
+        self.clear_cache()
 
     def __del__(self):
         #
@@ -490,11 +478,11 @@ class PluginEnvironment(object):
 
     def __contains__(self, cls):
         """
-        Return whether the given service is in the list of active services.
+        Return whether the given service is in the set of services.
         """
         return cls in self.services
 
-    def active_services(self, cls, all=False):
+    def active_services(self, cls, all=False, key=None):
         """
         Return the services that have been activated for a specific interface class.
         """
@@ -502,19 +490,32 @@ class PluginEnvironment(object):
             id = cls.__class__
         else:
             id = cls
-        if not issubclass(id,Interface):
-            raise PluginError, "PluginEnvironment[x] expects "+str(id)+" to be an Interface class"
-        #
-        # This could be optimized with a cache, but let's do that later...
-        #
-        return filter(lambda x: id in x.__interfaces__ and (all or x.enabled()), self.services)
+        try:
+            return self._cache[id,all,key]
+        except KeyError:
+            if not issubclass(id,Interface):
+                raise PluginError, "PluginEnvironment[x] expects "+str(id)+" to be an Interface class"
+            strkey = str(key)
+            tmp = filter(lambda x: id in x.__interfaces__ and (all or x.enabled()) and (key is None or x.name == strkey), self.services)
+            self._cache[id,all,key]=tmp
+            return tmp
 
     def activate(self, service):
         """
-        This inserts the service into the manager and the global registry.
+        This adds the service to this environment.
         """
         self.log.info("Adding service %s to environment %s" % (service.name, self.name))
         self.services.add(service)
+        self.clear_cache()
+
+    def deactivate(self, service):
+        """
+        This removes the service from this environment.
+        """
+        self.log.info("Removing service %s from environment %s" % (service.name, self.name))
+        if service in self.services:
+            self.services.remove(service)
+        self.clear_cache()
 
     def __repr__(self):
         self.pprint()
@@ -587,6 +588,11 @@ class PluginEnvironment(object):
 
         for loader in self.loaders:
             loader.load(self, search_path, disable_p, name_p)
+        self.clear_cache()
+
+    def clear_cache(self):
+        """ Clear the cache of active services """
+        self._cache = {}
 
 
 #
@@ -674,7 +680,22 @@ class PluginMeta(type):
             d['__singleton__'] = True
         else:
             d['__singleton__'] = False
-        new_class = type.__new__(cls, name, bases, d)
+        #
+        # Add interfaces to the list of base classes if they are
+        # declared inherited.
+        #
+        flag=False
+        bases = list(bases)
+        for interface in d.get('_inherited_interfaces', set()):
+            if not interface in bases:
+                bases.append(interface)
+                flag=True
+        if flag:
+            cls=MergedPluginMeta
+        #
+        # Create new class
+        #
+        new_class = type.__new__(cls, name, tuple(bases), d)
         if d['__singleton__']:
             #
             # Here, we create an instance of a singleton class, which
@@ -692,6 +713,11 @@ class PluginMeta(type):
         PluginGlobals.plugin_registry[name] = new_class
         return new_class
 
+class MergedPluginMeta(PluginMeta,InterfaceMeta):
+
+    def __new__(cls, name, bases, d):
+        return PluginMeta.__new__(cls, name, bases, d)
+
 
 class Plugin(object):
     """Base class for plugins.  A `service' is an instance of a Plugin.
@@ -718,10 +744,7 @@ class Plugin(object):
                 self = super(Plugin, cls).__new__(cls)
                 self.id=PluginGlobals.next_id()
                 self.name = self.__class__.__name__
-                for interface in self.__interfaces__:
-                    for ns in self.__interfaces__[interface]:
-                        PluginGlobals.env(ns).activate(self)
-                PluginGlobals.env(self.__plugin_namespace__).activate(self)
+                self.activate()
             self._enable = True
             cls.__instance__ = self
             return self
@@ -736,16 +759,17 @@ class Plugin(object):
         #
         # Activate the Plugin
         #
-        for interface in self.__interfaces__:
-            for ns in self.__interfaces__[interface]:
-                PluginGlobals.env(ns).activate(self)
-        PluginGlobals.env(self.__plugin_namespace__).activate(self)
+        self.activate()
         return self
 
-    def implements(interface, namespace=None):
+    def implements(interface, namespace=None, inherit=False):
         """
         Can be used in the class definition of `Plugin` subclasses to
-        declare the extension points that are implemented by this class.
+        declare the extension points that are implemented by this 
+        interface class.
+
+        If the `inherits` option is True, then this `Plugin` class
+        inherits from the `interface` class.
         """
         frame = sys._getframe(1)
         locals_ = frame.f_locals
@@ -759,6 +783,8 @@ class Plugin(object):
         #
         if namespace is None:
             namespace = interface.__interface_namespace__
+        if inherit:
+            locals_.setdefault('_inherited_interfaces', set()).add(interface)
         locals_.setdefault('_implements', {}).setdefault(interface,[]).append(namespace)
     implements = staticmethod(implements)
 
@@ -769,13 +795,47 @@ class Plugin(object):
         else:
             return '<Plugin %s %r>' % (self.__class__.__name__, self.name)
 
+    def activate(self):
+        """
+        Add this service to the global environment, and environments that manage the service's
+        interfaces.
+        """
+        for interface in self.__interfaces__:
+            for ns in self.__interfaces__[interface]:
+                PluginGlobals.env(ns).activate(self)
+        PluginGlobals.env(self.__plugin_namespace__).activate(self)
+        
+    def deactivate(self):
+        """
+        Remove this service from the global environment, and environments that manage the service's
+        interfaces.
+        """
+        for interface in self.__interfaces__:
+            for ns in self.__interfaces__[interface]:
+                PluginGlobals.env(ns).deactivate(self)
+        PluginGlobals.env(self.__plugin_namespace__).deactivate(self)
+        
     def disable(self):
         """Disable this plugin."""
         self._enable = False
+        #
+        # Clear the cache for environments that use this plugin
+        #
+        for interface in self.__interfaces__:
+            for ns in self.__interfaces__[interface]:
+                PluginGlobals.env(ns).clear_cache()
+        PluginGlobals.env(self.__plugin_namespace__).clear_cache()
 
     def enable(self):
         """Enable this plugin."""
         self._enable = True
+        #
+        # Clear the cache for environments that use this plugin
+        #
+        for interface in self.__interfaces__:
+            for ns in self.__interfaces__[interface]:
+                PluginGlobals.env(ns).clear_cache()
+        PluginGlobals.env(self.__plugin_namespace__).clear_cache()
 
     def enabled(self):
         """Can be overriden to control whether a plugin is enabled."""
